import tkinter as tk
from functools import partial
from logging import getLogger
from os import path, removedirs, chdir, walk, remove, sep
from pathlib import Path
from shutil import move, copy2, rmtree
from sys import platform
from time import time
from tkinter import filedialog, messagebox

from moht import PLUGINS2CLEAN, VERSION, TES3CMD
from moht.utils import is_latest_ver, parse_cleaning, run_cmd, here


class MohtTkGui(tk.Frame):
    def __init__(self, master: tk.Tk,) -> None:
        """
        Create basic GUI for Mod Helper Tool application.

        :param master: Top level widget
        """
        super().__init__(master)
        self.logger = getLogger(__name__)
        self.master = master
        self.statusbar = tk.StringVar()
        self._mods_dir = tk.StringVar()
        self._morrowind_dir = tk.StringVar()
        self._tes3cmd = tk.StringVar()
        self.chkbox_backup = tk.BooleanVar()
        self.chkbox_cache = tk.BooleanVar()
        self.rb_tes3cmd = tk.StringVar()
        self.stats = {'all': 0, 'cleaned': 0, 'clean': 0, 'error': 0, 'time': 0.0}
        self._init_widgets()
        self.statusbar.set(f'ver. {VERSION}')
        self._mods_dir.set(str(Path.home()))
        self._morrowind_dir.set(str(Path.home()))
        self._tes3cmd.set(path.join(here(__file__), 'resources', TES3CMD[platform]['0_37']))
        self.chkbox_backup.set(True)
        self.chkbox_cache.set(True)
        self.tes3cmd_file.config(state=tk.DISABLED)
        self.tes3cmd_btn.config(state=tk.DISABLED)

    def _init_widgets(self) -> None:
        self.master.columnconfigure(index=0, weight=10)
        self.master.rowconfigure(index=0, weight=1)
        self.master.rowconfigure(index=1, weight=1)
        self.master.rowconfigure(index=2, weight=1)

        mods_dir = tk.Entry(master=self.master, textvariable=self._mods_dir)
        morrowind_dir = tk.Entry(master=self.master, textvariable=self._morrowind_dir)
        self.tes3cmd_file = tk.Entry(master=self.master, textvariable=self._tes3cmd)
        mods_btn = tk.Button(master=self.master, text='Select Mods', width=16, command=partial(self.select_dir, self._mods_dir))
        morrowind_btn = tk.Button(master=self.master, text='Select Data Files', width=16, command=partial(self.select_dir, self._morrowind_dir))
        self.tes3cmd_btn = tk.Button(master=self.master, text='Select tes3cmd', width=16, command=partial(self.select_tes3cmd_file, self._tes3cmd))
        self.clean_btn = tk.Button(master=self.master, text='Clean Mods', width=16, command=self.start_clean)
        self.report_btn = tk.Button(master=self.master, text='Report', width=16, state=tk.DISABLED, command=self.report)
        chkupd_btn = tk.Button(master=self.master, text='Check Updates', width=16, command=self.check_updates)
        close_btn = tk.Button(master=self.master, text='Close Tool', width=16, command=self.master.destroy)
        statusbar = tk.Label(master=self.master, textvariable=self.statusbar)
        chkbox_frame = tk.LabelFrame(master=self.master, text='After successful clean-up:', relief=tk.GROOVE, borderwidth=2)
        chkbox_backup = tk.Checkbutton(master=chkbox_frame, text='Remove backups of plugins', variable=self.chkbox_backup)
        chkbox_cache = tk.Checkbutton(master=chkbox_frame, text='Remove cache of master files', variable=self.chkbox_cache)

        rb_frame = tk.LabelFrame(master=self.master, text='Version of tes3cmd:', relief=tk.GROOVE, borderwidth=2)
        rb_037 = tk.Radiobutton(master=rb_frame, text='built-in v0.37', value='0_37', variable=self.rb_tes3cmd, command=self._rb_tes3cmd_toggled)
        rb_040 = tk.Radiobutton(master=rb_frame, text='built-in v0.40', value='0_40', variable=self.rb_tes3cmd, command=self._rb_tes3cmd_toggled)
        rb_custom = tk.Radiobutton(master=rb_frame, text='custom', value='custom', variable=self.rb_tes3cmd, command=self._rb_tes3cmd_toggled)
        rb_037.select()

        mods_dir.grid(row=0, column=0, padx=2, pady=2, columnspan=2, sticky=f'{tk.W}{tk.E}')
        morrowind_dir.grid(row=1, column=0, padx=2, pady=2, columnspan=2, sticky=f'{tk.W}{tk.E}')
        self.tes3cmd_file.grid(row=2, column=0, padx=2, pady=2, columnspan=2, sticky=f'{tk.W}{tk.E}')
        chkbox_frame.grid(row=3, column=0, padx=2, pady=2, rowspan=2, sticky=tk.W)
        chkbox_backup.grid(row=3, column=0, padx=2, pady=2, sticky=tk.W)
        chkbox_cache.grid(row=4, column=0, padx=2, pady=2, sticky=tk.W)
        statusbar.grid(row=7, column=0, columnspan=3, sticky=tk.W)

        rb_frame.grid(row=3, column=1, padx=2, pady=2, rowspan=3)
        rb_037.grid(row=3, column=1, padx=2, pady=2, sticky=tk.W)
        rb_040.grid(row=4, column=1, padx=2, pady=2, sticky=tk.W)
        rb_custom.grid(row=5, column=1, padx=2, pady=2, sticky=tk.W)

        mods_btn.grid(row=0, column=2, padx=2, pady=2)
        morrowind_btn.grid(row=1, column=2, padx=2, pady=2)
        self.tes3cmd_btn.grid(row=2, column=2, padx=2, pady=2)
        self.clean_btn.grid(row=3, column=2, padx=2, pady=2)
        self.report_btn.grid(row=4, column=2, padx=2, pady=2)
        chkupd_btn.grid(row=5, column=2, padx=2, pady=2)
        close_btn.grid(row=6, column=2, padx=2, pady=2)

    def select_dir(self, text_var: tk.StringVar) -> None:
        """
        Select directory location.

        :param text_var: StringVar of Entry to update
        """
        directory = filedialog.askdirectory(initialdir=str(Path.home()), title='Select directory')
        self.logger.debug(f'Directory: {directory}')
        text_var.set(f'{directory}')

    def select_tes3cmd_file(self, text_var: tk.StringVar) -> None:
        """
        Select tes3cmd file location.

        :param text_var: StringVar of Entry to update
        """
        filename = filedialog.askopenfilename(initialdir=str(Path.home()), title='Select file')
        self.logger.debug(f'File: {filename}')
        text_var.set(f'{filename}')
        if self._check_clean_bin():
            self.clean_btn.config(state=tk.ACTIVE)
        else:
            self.clean_btn.config(state=tk.DISABLED)

    def _rb_tes3cmd_toggled(self) -> None:
        if self.rb_tes3cmd.get() == 'custom':
            self.tes3cmd_file.config(state=tk.NORMAL)
            self.tes3cmd_btn.config(state=tk.NORMAL)
        else:
            self.logger.debug(f'tes3cmd version: {self.rb_tes3cmd.get()}')
            self._tes3cmd.set(path.join(here(__file__), 'resources', TES3CMD[platform][self.rb_tes3cmd.get()]))
            self.tes3cmd_file.config(state=tk.DISABLED)
            self.tes3cmd_btn.config(state=tk.DISABLED)

    def start_clean(self) -> None:
        """Start cleaning process."""
        if not all([path.isdir(folder) for folder in [self.mods_dir, self.morrowind_dir]]):
            self.statusbar.set('Check directories and try again')
            return
        all_plugins = [Path(path.join(root, filename))
                       for root, _, files in walk(self.mods_dir)
                       for filename in files
                       if filename.lower().endswith('.esp') or filename.lower().endswith('.esm')]
        self.logger.debug(f'all_plugins: {len(all_plugins)}: {all_plugins}')
        plugins_to_clean = [plugin_file for plugin_file in all_plugins if str(plugin_file).split(sep)[-1] in PLUGINS2CLEAN]
        no_of_plugins = len(plugins_to_clean)
        self.logger.debug(f'to_clean: {no_of_plugins}: {plugins_to_clean}')
        chdir(self.morrowind_dir)
        self.stats = {'all': no_of_plugins, 'cleaned': 0, 'clean': 0, 'error': 0}
        start = time()
        for idx, plug in enumerate(plugins_to_clean, 1):
            self.logger.debug(f'---------------------------- {idx} / {no_of_plugins} ---------------------------- ')
            self.logger.debug(f'Copy: {plug} -> {self.morrowind_dir}')
            copy2(plug, self.morrowind_dir)
            mod_file = str(plug).split(sep)[-1]
            out, err = run_cmd(f'{self.tes3cmd} clean --output-dir --overwrite "{mod_file}"')
            result, reason = parse_cleaning(out, err, mod_file)
            self.logger.debug(f'Result: {result}, Reason: {reason}')
            self._update_stats(mod_file, plug, reason, result)
            if self.chkbox_backup.get():
                self.logger.debug(f'Remove: {self.morrowind_dir}/{mod_file}')
                remove(f'{self.morrowind_dir}/{mod_file}')
        self.logger.debug(f'---------------------------- Done: {no_of_plugins} ---------------------------- ')
        if self.chkbox_cache.get():
            removedirs(f'{self.morrowind_dir}/1')
            cachedir = 'tes3cmd' if platform == 'win32' else '.tes3cmd-3'
            rmtree(f'{self.morrowind_dir}/{cachedir}', ignore_errors=True)
        cleaning_time = time() - start
        self.stats['time'] = cleaning_time
        self.logger.debug(f'Total time: {cleaning_time} s')
        self.statusbar.set('Done. See report!')
        self.report_btn.config(state=tk.NORMAL)

    def _update_stats(self, mod_file: str, plug: Path, reason: str, result: bool) -> None:
        if result:
            self.logger.debug(f'Move: {self.morrowind_dir}/1/{mod_file} -> {plug}')
            move(f'{self.morrowind_dir}/1/{mod_file}', plug)
            self.stats['cleaned'] += 1
        if not result and reason == 'not modified':
            self.stats['clean'] += 1
        if not result and 'not found' in reason:
            self.stats['error'] += 1
            esm = self.stats.get(reason, 0)
            esm += 1
            self.stats.update({reason: esm})

    def report(self) -> None:
        """Show report after clean-up."""
        self.logger.debug(f'Report: {self.stats}')
        report = f'Detected plugins to clean: {self.stats["all"]}\n'
        report += f'Already clean plugins: {self.stats["clean"]}\n'
        report += f'Cleaned plugins: {self.stats["cleaned"]}\n'
        report += '\n'.join([f'Error {k}: {self.stats[k]}' for k in self.stats if 'not found' in k])
        report += '\n\nCopy missing esm file(s) to Data Files directory and clean again.\n\n' if 'Error' in report else '\n'
        report += f'Total time: {self.stats["time"]:.2f} s'
        messagebox.showinfo('Cleaning Report', report)
        self.report_btn.config(state=tk.DISABLED)
        self.statusbar.set(f'ver. {VERSION}')

    def check_updates(self):
        """Check for updates."""
        latest, desc = is_latest_ver(package='moht', current_ver=VERSION)
        current_ver = 'No updates' if latest else f'Update available: {desc}'
        self.statusbar.set(f'ver. {VERSION} - {current_ver}')

    def _check_clean_bin(self) -> bool:
        self.logger.debug('Checking tes3cmd')
        out, err = run_cmd(f'{self.tes3cmd} -h')
        result, reason = parse_cleaning(out, err, '')
        self.logger.debug(f'Result: {result}, Reason: {reason}')
        if not result:
            self.statusbar.set(f'Error: {reason}')
            if 'Config::IniFiles' in reason:
                reason = 'Use package manager, check for `perl-Config-IniFiles` or a similar package.\n\nOr run from a terminal:\ncpan install Config::IniFiles'
            elif 'Not tes3cmd' in reason:
                reason = 'Selected file is not a valid tes3cmd executable.\n\nPlease select a correct binary file.'
            messagebox.showerror('Not tes3cmd', reason)
        return result

    @property
    def mods_dir(self) -> str:
        """
        Get root of mods directory.

        :return: mods dir as string
        """
        return str(self._mods_dir.get())

    @mods_dir.setter
    def mods_dir(self, value: Path) -> None:
        self._mods_dir.set(str(value))

    @property
    def morrowind_dir(self) -> str:
        """
        Get Morrowind Data Files directory.

        :return: morrowind dir as string
        """
        return str(self._morrowind_dir.get())

    @morrowind_dir.setter
    def morrowind_dir(self, value: Path) -> None:
        self._morrowind_dir.set(str(value))

    @property
    def tes3cmd(self) -> str:
        """
        Get tes3cmd binary file path.

        :return: tes3cmd file as string
        """
        return str(self._tes3cmd.get())

    @tes3cmd.setter
    def tes3cmd(self, value: Path) -> None:
        self._tes3cmd.set(str(value))



#!/usr/bin/env python
import signal
import sys
from argparse import ArgumentParser, RawTextHelpFormatter, Namespace
from logging import getLogger
from os import name
from os import path
from platform import architecture, uname, python_implementation, python_version
from sys import platform
from tempfile import gettempdir

from moht import VERSION
from moht.log import config_logger
from moht.utils import here


def run_tk():
    """Function to start Mod Helper Tool Tk GUI."""
    import tkinter
    from moht.tkgui import MohtTkGui

    root = tkinter.Tk()
    width, height = 500, 280
    root.title('Mod Helper Tool')
    root.geometry(f'{width}x{height}')
    root.minsize(width=width, height=height)
    root.iconphoto(False, tkinter.PhotoImage(file=path.join(here(__file__), 'img', 'moht.png')))
    window = MohtTkGui(master=root)
    window.mainloop()


def run_qt():
    """Function to start Mod Helper Tool QtGUI."""
    from PyQt5 import QtCore
    from PyQt5.QtCore import QLibraryInfo, QTranslator, QLocale
    from PyQt5.QtWidgets import QApplication
    from moht.qtgui import MohtQtGui

    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app = QApplication(sys.argv)

    translator = QTranslator(app)
    if translator.load(QLocale.system(), 'qtbase', '_', QLibraryInfo.location(QLibraryInfo.TranslationsPath)):
        app.installTranslator(translator)
    translator = QTranslator(app)
    if translator.load(QLocale.system(), 'qtgui', '-', here(__file__)):  # change to _
        app.installTranslator(translator)

    window = MohtQtGui()
    window.show()
    sys.exit(app.exec())


def run(cli_opts: Namespace) -> None:
    """
    Function to start selected GUI of Mod Helper Tool.

    :param cli_opts: arguments from CLI
    """
    config_logger(verbose=cli_opts.verbose, quiet=cli_opts.quiet)
    logger = getLogger(f'moht.{__name__}')
    logger.info(f'Log file stored at: {path.join(gettempdir(), "moht.log")}')
    logger.info(f'moht v{VERSION} https://gitlab.com/modding-openmw/modhelpertool')
    logger.debug(f'Arch: {name} / {platform} / {" / ".join(architecture())}')
    logger.debug(f'Python: {python_implementation()}-{python_version()}')
    logger.debug(f'{uname()}')
    globals()[f'run_{cli_opts.gui}']()


if __name__ == '__main__':
    parser = ArgumentParser(description='Simple yet powerful tool to help you manage your mods in several ways.', formatter_class=RawTextHelpFormatter)
    parser.add_argument('-V', '--version', action='version', version='%(prog)s Version: ' + VERSION)
    parser.add_argument('-q', '--quiet', action='store_true', dest='quiet', default=False, help='be quiet')
    parser.add_argument('-v', '--verbose', action='count', dest='verbose', default=0, help='increase output verbosity')
    gui = parser.add_subparsers(title='gui', dest='gui', description='Available subcommands', help='choose one of GUI')
    gui_qt = gui.add_parser(name='qt', help='starting Qt5 GUI interface for Moht', formatter_class=RawTextHelpFormatter)
    gui_qt.add_argument('-style', dest='style', help='style for QtGUI: "fusion" (default) or "windows".', default='fusion')
    gui_tk = gui.add_parser(name='tk', help='starting Tk GUI interface for Moht', formatter_class=RawTextHelpFormatter)
    args = parser.parse_args()
    if args.gui:
        run(args)
    else:
        parser.print_help()



from datetime import datetime
from logging import DEBUG, StreamHandler, INFO, Formatter, WARNING, getLogger
from logging.handlers import RotatingFileHandler
from os import path
from tempfile import gettempdir


__logger_configured = False


def _configure_logger(verbose: int, quiet: bool) -> None:
    root_logger = getLogger('moht')
    root_logger.setLevel(DEBUG)
    file_hand = RotatingFileHandler(filename=path.join(gettempdir(), 'moht.log'), mode='a', encoding='utf-8', maxBytes=5 * 1024 * 1024, backupCount=1)
    file_hand.setLevel(DEBUG)
    file_hand.setFormatter(Formatter('%(asctime)s | %(name)-17s | %(levelname)-7s | %(threadName)-10s | %(message)s / %(funcName)s:%(lineno)d'))
    root_logger.addHandler(file_hand)
    header = '#' * 60
    root_logger.debug(f'\n{header}\nStart session: {datetime.now()}\n{header}')

    if not quiet:
        stream_hand = StreamHandler()
        if verbose == 0:
            stream_hand.setLevel(WARNING)
        if verbose == 1:
            stream_hand.setLevel(INFO)
        if verbose >= 2:
            stream_hand.setLevel(DEBUG)
        stream_hand.setFormatter(Formatter('%(levelname)-7s | %(message)s'))
        root_logger = getLogger('moht')
        root_logger.addHandler(stream_hand)

    global __logger_configured
    __logger_configured = True


def config_logger(verbose: int, quiet: bool) -> None:
    """
    Configure global logger, handlers and set formatters.

    :param verbose: increase verbosity with increment of integer
    :param quiet: turn on/off quiet mode
    """
    if not __logger_configured:
        _configure_logger(verbose, quiet)



import webbrowser
from functools import partial
from logging import getLogger
from os import path, removedirs, chdir, walk, remove, sep
from pathlib import Path
from shutil import move, copy2, rmtree
from sys import version_info, platform
from tempfile import gettempdir
from time import time
from typing import Optional

import qtawesome
from PyQt5 import QtCore, uic
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QMainWindow, QMessageBox, QDialog, QFileDialog

from moht import PLUGINS2CLEAN, VERSION, TES3CMD, qtgui_rc
from moht.utils import parse_cleaning, run_cmd, is_latest_ver, here

res = qtgui_rc  # prevent to remove import statement accidentally


def tr(text2translate: str):
    """
    Translate wrapper function.

    :param text2translate: string to translate
    :return:
    """
    # return QtCore.QCoreApplication.translate('mw_gui', text2translate)
    return QtCore.QCoreApplication.translate('@default', text2translate)


class MohtQtGui(QMainWindow):
    def __init__(self) -> None:
        """Mod Helper Tool Qt5 GUI."""
        super(MohtQtGui, self).__init__(flags=QtCore.Qt.Window)
        self.logger = getLogger(__name__)
        uic.loadUi(f'{here(__file__)}/ui/qtgui.ui', self)
        self.threadpool = QtCore.QThreadPool.globalInstance()
        self.logger.debug(f'QThreadPool with {self.threadpool.maxThreadCount()} thread(s)')
        self._le_status = {'le_mods_dir': False, 'le_morrowind_dir': False, 'le_tes3cmd': False}
        self.stats = {'all': 0, 'cleaned': 0, 'clean': 0, 'error': 0, 'time': 0.0}
        self.tes3cmd = TES3CMD[platform]['0_37']
        self._init_menu_bar()
        self._init_buttons()
        self._init_radio_buttons()
        self._init_line_edits()
        self.statusbar.showMessage(f'ver. {VERSION}')
        self._set_icons()

    def _init_menu_bar(self) -> None:
        self.actionQuit.triggered.connect(self.close)
        self.actionAboutMoht.triggered.connect(AboutDialog(self).open)
        self.actionAboutQt.triggered.connect(partial(self._show_message_box, kind_of='aboutQt', title='About Qt'))
        self.actionReportIssue.triggered.connect(self._report_issue)
        self.actionCheckUpdates.triggered.connect(self._check_updates)

    def _init_buttons(self) -> None:
        self.pb_mods_dir.clicked.connect(partial(self._run_file_dialog, for_load=True, for_dir=True, widget_name='le_mods_dir'))
        self.pb_morrowind_dir.clicked.connect(partial(self._run_file_dialog, for_load=True, for_dir=True, widget_name='le_morrowind_dir'))
        self.pb_tes3cmd.clicked.connect(partial(self._run_file_dialog, for_load=True, for_dir=False, widget_name='le_tes3cmd'))
        self.pb_clean.clicked.connect(self._pb_clean_clicked)
        self.pb_report.clicked.connect(self._pb_report_clicked)
        self.pb_chk_updates.clicked.connect(self._check_updates)

    def _init_line_edits(self):
        self.le_mods_dir.textChanged.connect(partial(self._is_dir_exists, widget_name='le_mods_dir'))
        self.le_morrowind_dir.textChanged.connect(partial(self._is_dir_exists, widget_name='le_morrowind_dir'))
        self.le_tes3cmd.textChanged.connect(partial(self._is_file_exists, widget_name='le_tes3cmd'))
        self._set_le_tes3cmd()
        self.le_mods_dir.setText(str(Path.home()))
        self.le_morrowind_dir.setText(str(Path.home()))

    def _init_radio_buttons(self):
        for ver in ['0_37', '0_40']:
            getattr(self, f'rb_{ver}').toggled.connect(partial(self._rb_tes3cmd_toggled, ver))

    def _rb_tes3cmd_toggled(self, version: str) -> None:
        self.tes3cmd = TES3CMD[platform][version]
        self._set_le_tes3cmd()

    def _pb_clean_clicked(self) -> None:
        all_plugins = [Path(path.join(root, filename))
                       for root, _, files in walk(self.le_mods_dir.text())
                       for filename in files
                       if filename.lower().endswith('.esp') or filename.lower().endswith('.esm')]
        self.logger.debug(f'all_plugins: {len(all_plugins)}: {all_plugins}')
        plugins_to_clean = [plugin_file for plugin_file in all_plugins if str(plugin_file).split(sep)[-1] in PLUGINS2CLEAN]
        no_of_plugins = len(plugins_to_clean)
        self.logger.debug(f'to_clean: {no_of_plugins}: {plugins_to_clean}')
        chdir(self.le_morrowind_dir.text())
        self.stats = {'all': no_of_plugins, 'cleaned': 0, 'clean': 0, 'error': 0}
        start = time()
        for idx, plug in enumerate(plugins_to_clean, 1):
            self.logger.debug(f'---------------------------- {idx} / {no_of_plugins} ---------------------------- ')
            self.logger.debug(f'Copy: {plug} -> {self.le_morrowind_dir.text()}')
            copy2(plug, self.le_morrowind_dir.text())
            mod_file = str(plug).split(sep)[-1]
            out, err = run_cmd(f'{self.le_tes3cmd.text()} clean --output-dir --overwrite "{mod_file}"')
            result, reason = parse_cleaning(out, err, mod_file)
            self.logger.debug(f'Result: {result}, Reason: {reason}')
            self._update_stats(mod_file, plug, reason, result)
            if self.cb_rm_bakup.isChecked():
                self.logger.debug(f'Remove: {self.le_morrowind_dir.text()}/{mod_file}')
                remove(f'{self.le_morrowind_dir.text()}/{mod_file}')
        self.logger.debug(f'---------------------------- Done: {no_of_plugins} ---------------------------- ')
        if self.cb_rm_cache.isChecked():
            removedirs(f'{self.le_morrowind_dir.text()}/1')
            cachedir = 'tes3cmd' if platform == 'win32' else '.tes3cmd-3'
            rmtree(f'{self.le_morrowind_dir.text()}/{cachedir}', ignore_errors=True)
        cleaning_time = time() - start
        self.stats['time'] = cleaning_time
        self.logger.debug(f'Total time: {cleaning_time} s')
        self.statusbar.showMessage('Done. See report!')
        self.pb_report.setEnabled(True)

    def _pb_report_clicked(self) -> None:
        """Show report after clean-up."""
        self.logger.debug(f'Report: {self.stats}')
        report = f'Detected plugins to clean: {self.stats["all"]}\n'
        report += f'Already clean plugins: {self.stats["clean"]}\n'
        report += f'Cleaned plugins: {self.stats["cleaned"]}\n'
        report += '\n'.join([f'Error {k}: {self.stats[k]}' for k in self.stats if 'not found' in k])
        report += '\n\nCopy missing esm file(s) to Data Files directory and clean again.\n\n' if 'Error' in report else '\n'
        report += f'Total time: {self.stats["time"]:.2f} s'
        self._show_message_box(kind_of='information', title='Cleaning Report', message=report)
        self.pb_report.setEnabled(False)
        self.statusbar.showMessage(f'ver. {VERSION}')

    def _check_updates(self):
        latest, desc = is_latest_ver(package='moht', current_ver=VERSION)
        current_ver = 'No updates' if latest else f'Update available: {desc}'
        self.statusbar.showMessage(f'ver. {VERSION} - {current_ver}')

    def _set_le_tes3cmd(self) -> None:
        self.le_tes3cmd.setText(path.join(here(__file__), 'resources', self.tes3cmd))

    def _update_stats(self, mod_file: str, plug: Path, reason: str, result: bool) -> None:
        if result:
            self.logger.debug(f'Move: {self.le_morrowind_dir.text()}/1/{mod_file} -> {plug}')
            move(f'{self.le_morrowind_dir.text()}/1/{mod_file}', plug)
            self.stats['cleaned'] += 1
        if not result and reason == 'not modified':
            self.stats['clean'] += 1
        if not result and 'not found' in reason:
            self.stats['error'] += 1
            esm = self.stats.get(reason, 0)
            esm += 1
            self.stats.update({reason: esm})

    def _is_dir_exists(self, text: str, widget_name: str) -> None:
        dir_exists = path.isdir(text)
        self.logger.debug(f'Path: {text} for {widget_name} exists: {dir_exists}')
        self._line_edit_handling(widget_name, dir_exists)

    def _is_file_exists(self, text: str, widget_name) -> None:
        file_exists = path.isfile(text)
        self.logger.debug(f'Path: {text} for {widget_name} exists: {file_exists}')
        self._line_edit_handling(widget_name, file_exists)

    def _line_edit_handling(self, widget_name: str, path_exists: bool) -> None:
        """
        Mark text of LieEdit as red if path does not exist.

        Additionally, save status and enable /disable Clean button base on it.

        :param widget_name: widget name
        :param path_exists: bool for path existence
        """
        self._le_status[widget_name] = path_exists
        if path_exists and widget_name == 'le_tes3cmd':
            getattr(self, widget_name).setStyleSheet('')
            self._le_status[widget_name] = self._check_clean_bin()
        elif path_exists and widget_name != 'le_tes3cmd':
            getattr(self, widget_name).setStyleSheet('')
        else:
            getattr(self, widget_name).setStyleSheet('color: red;')
        if all(self._le_status.values()):
            self.pb_clean.setEnabled(True)
        else:
            self.pb_clean.setEnabled(False)

    def _check_clean_bin(self) -> bool:
        self.logger.debug('Checking tes3cmd')
        out, err = run_cmd(f'{self.le_tes3cmd.text()} -h')
        result, reason = parse_cleaning(out, err, '')
        self.logger.debug(f'Result: {result}, Reason: {reason}')
        if not result:
            self.statusbar.showMessage(f'Error: {reason}')
            if 'Config::IniFiles' in reason:
                reason = 'Use package manager, check for `perl-Config-IniFiles` or a similar package.\n\nOr run from a terminal:\ncpan install Config::IniFiles'
            elif 'Not tes3cmd' in reason:
                reason = 'Selected file is not a valid tes3cmd executable.\n\nPlease select a correct binary file.'
            self._show_message_box(kind_of='warning', title='Not tes3cmd', message=reason)
        return result

    def _set_icons(self, button: Optional[str] = None, icon_name: Optional[str] = None, color: str = 'black', spin: bool = False):
        """
        Universal method to set icon for QPushButtons.

        When button is provided without icon_name, current button icon will be removed.
        When none of button nor icon_name are provided, default starting icons are set for all buttons.

        :param button: button name
        :param icon_name: ex: spinner, check, times, pause
        :param color: ex: red, green, black
        :param spin: spinning icon: True or False
        """
        if not (button or icon_name):
            self.pb_mods_dir.setIcon(qtawesome.icon('fa5s.folder', color='brown'))
            self.pb_morrowind_dir.setIcon(qtawesome.icon('fa5s.folder', color='brown'))
            self.pb_tes3cmd.setIcon(qtawesome.icon('fa5s.file', color='brown'))
            self.pb_clean.setIcon(qtawesome.icon('fa5s.hand-sparkles', color='brown'))
            self.pb_report.setIcon(qtawesome.icon('fa5s.file-contract', color='brown'))
            self.pb_chk_updates.setIcon(qtawesome.icon('fa5s.arrow-down', color='brown'))
            self.pb_close.setIcon(qtawesome.icon('fa5s.sign-out-alt', color='brown'))
            return
        btn = getattr(self, button)  # type: ignore
        if spin and icon_name:
            icon = qtawesome.icon('{}'.format(icon_name), color=color, animation=qtawesome.Spin(btn, 2, 1))
        elif not spin and icon_name:
            icon = qtawesome.icon('{}'.format(icon_name), color=color)
        else:
            icon = QIcon()
        btn.setIcon(icon)

    def _run_file_dialog(self, for_load: bool, for_dir: bool, widget_name: Optional[str] = None, file_filter: str = 'All Files [*.*](*.*)') -> str:
        """
        Handling open/save dialog to select file or folder.

        :param for_load: if True show window for load, for save otherwise
        :param for_dir: if True show window for selecting directory only, if False selectting file only
        :param file_filter: list of types of files ;;-seperated: Text [*.txt](*.txt)
        :return: full path to file or directory
        """
        result_path = ''
        if file_filter != 'All Files [*.*](*.*)':
            file_filter = '{};;All Files [*.*](*.*)'.format(file_filter)
        if for_load and for_dir:
            result_path = QFileDialog.getExistingDirectory(QFileDialog(), caption='Open Directory', directory=str(Path.home()),
                                                           options=QFileDialog.ShowDirsOnly)
        if for_load and not for_dir:
            result_path = QFileDialog.getOpenFileName(QFileDialog(), caption='Open File', directory=str(Path.home()),
                                                      filter=file_filter, options=QFileDialog.ReadOnly)
            result_path = result_path[0]
        if not for_load and not for_dir:
            result_path = QFileDialog.getSaveFileName(QFileDialog(), caption='Save File', directory=str(Path.home()),
                                                      filter=file_filter, options=QFileDialog.ReadOnly)
            result_path = result_path[0]
        if widget_name is not None:
            getattr(self, widget_name).setText(result_path)
        return result_path

    def _show_message_box(self, kind_of: str, title: str, message: str = '') -> None:
        """
        Generic method to show any QMessageBox delivered with Qt.

        :param kind_of: any of: information, question, warning, critical, about or aboutQt
        :param title: Title of modal window
        :param message: text of message, default is empty
        """
        message_box = getattr(QMessageBox, kind_of)
        if kind_of == 'aboutQt':
            message_box(self, title)
        else:
            message_box(self, title, message)

    @staticmethod
    def _report_issue():
        webbrowser.open('https://gitlab.com/modding-openmw/modhelpertool/issues', new=2)


class AboutDialog(QDialog):
    def __init__(self, parent) -> None:
        """Moht about dialog window."""
        super(AboutDialog, self).__init__(parent)
        uic.loadUi(f'{here(__file__)}/ui/about.ui', self)
        self.setup_text()

    def setup_text(self) -> None:
        """Prepare text information about Moht application."""
        qt_version = f'{QtCore.PYQT_VERSION_STR} / <b>Qt</b>: {QtCore.QT_VERSION_STR}'
        log_path = path.join(gettempdir(), 'moht.log')
        text = self.label.text().rstrip('</body></html>')
        text += f'<p>Attach log file: {log_path}<br/><br/>'
        text += f'<b>moht:</b> {VERSION}'
        text += '<br><b>python:</b> {0}.{1}.{2}-{3}.{4}'.format(*version_info)
        text += f'<br><b>PyQt:</b> {qt_version}</p></body></html>'
        self.label.setText(text)



from logging import getLogger
from os import linesep, path
from re import search, MULTILINE
from shlex import split
from subprocess import Popen, PIPE
from sys import platform
from typing import Tuple

from packaging import version

logger = getLogger(__name__)


def run_cmd(cmd: str) -> Tuple[str, str]:
    """
    Run command and return stdout and stderr.

    :param cmd: command to execute
    :return: stdout, stderr
    """
    cmd2exec = split(cmd) if platform == 'linux' else cmd
    logger.debug(f'CMD: {cmd2exec}')
    stdout, stderr = Popen(cmd2exec, stdout=PIPE, stderr=PIPE).communicate()
    out, err = stdout.decode('utf-8'), stderr.decode('utf-8')
    logger.debug(f'StdOut: {out}')
    logger.debug(f'StdErr: {err}')
    return out, err


def parse_cleaning(out: str, err: str, mod_filename: str) -> Tuple[bool, str]:  # type: ignore
    """
    Parse output of cleaning command printout.

    :param out: Command STANDARD OUTPUT
    :param err: Command STANDARD ERROR
    :param mod_filename: Mod filename
    :return: Result and reason
    """
    ceases = {
        1: {'args': (r'\[ERROR \({}\): Master: (.* not found) in <DATADIR>]'.format(mod_filename), err, MULTILINE),
            'result': False},
        2: {'args': (r'{} was (not modified)'.format(mod_filename), out, MULTILINE),
            'result': False},
        3: {'args': (r'Output (saved) in: "1/{}"{}Original unaltered: "{}"'.format(mod_filename, linesep, mod_filename), out, MULTILINE),
            'result': True},
        4: {'args': (r'Can\'t locate Config/IniFiles.pm in @INC \(you may need to install the (Config::IniFiles module)\)', err, MULTILINE),
            'result': False},
        5: {'args': (r'(Usage): tes3cmd COMMAND OPTIONS plugin...', err, MULTILINE),
            'result': True},
    }
    for data in ceases.values():
        match = search(*data['args'])  # type: ignore
        if match:
            return data['result'], match.group(1)  # type: ignore
    return False, 'Not tes3cmd'


def is_latest_ver(package: str, current_ver: str) -> Tuple[bool, str]:
    """
    Check if installed package is the latest.

    :param package: package name
    :param current_ver: currently installed version
    """
    extra_data = current_ver
    out, err = run_cmd(f'pip install --dry-run --no-color --timeout 3 --retries 1 --progress-bar off --upgrade {package}')
    match = search(r'Would install\s.*{}-([\d.-]+)'.format(package), out)
    if match:
        extra_data = match.group(1)
        logger.debug(f'Latest available version: {extra_data}')
    match = search(r'no such option:\s(.*)', err)
    if match:
        extra_data = match.group(1)
        logger.warning(f'Version check failed, unknown switch: {extra_data}')
        out, _ = run_cmd('pip list')
        match = search(r'pip\s*([\d.]*)', out)
        if match:
            extra_data = f'unknown switch {extra_data} pip: {match.group(1)}'
            logger.debug(f'Pip version: {match.group(1)}')
    latest = _compare_versions(package, current_ver, extra_data)
    return latest, extra_data


def _compare_versions(package: str, current_ver: str, remote_ver: str) -> bool:
    """
    Compare versions.

    :param package:
    :param current_ver:
    :param remote_ver:
    :return:
    """
    latest = False
    if version.parse(remote_ver) > version.parse(current_ver):
        logger.info(f'There is new version of {package}: {remote_ver}')
    elif version.parse(remote_ver) <= version.parse(current_ver):
        logger.info(f'{package} is up-to-date version: {current_ver}')
        latest = True
    return latest


def here(filename: str) -> str:
    """
    Absolute path to directory contains filename.

    :param filename:
    :return: path to directory as string
    """
    return path.abspath(path.dirname(filename))
